
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"

#ifdef	G2T_RGB
#include "sif.h"
#include "sif_rgb.h"

#define	CM	unsigned char
#else
#include "sif_cm.h"

#define	CM	unsigned short
#endif

typedef	unsigned char	uChar;
typedef	unsigned short	uShort;

static void	Error(msg)
char		*msg;
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

static FILE	*file;

static int	Get()
{
	int	byte=fgetc(file);

	if (byte==EOF) Error("byte read error.");

	return byte;
}

static void	Skip(count)
int		count;
{
	while (--count>=0) (void)Get();
}

#ifdef	G2T_RGB
static void	ReadCM(rgbSize,BPS,R,G,B)
#else
static int	ReadCM(rgbSize,BPS,R,G,B)
#endif
int		rgbSize,BPS;
CM		*R,*G,*B;
{
	int	z,size=1<<BPS;
	double	ratio=(double)((1<<(8*sizeof(CM)))-1)/(double)((1<<rgbSize)-1);

	for (z=0; z<size; z++) {
	    R[z]=(CM)((double)Get()*ratio+0.5);
	    G[z]=(CM)((double)Get()*ratio+0.5);
	    B[z]=(CM)((double)Get()*ratio+0.5);
	}
	for (; z<256; z++) R[z]=G[z]=B[z]=0;
#ifndef	G2T_RGB
	return (BPS<=1)?1:(BPS<=2)?2:(BPS<=4)?4:8;
#endif
}

static int	ReadBlock(byte)
uChar		*byte;
{
	size_t	size=(size_t)Get();

	if (size!=0) {
	    if (size>255) Error("too large data block.");

	    if (fread(byte,sizeof(uChar),size,file)!=size)
		Error("data block read error.");
	}
	return (int)size;
}

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size_t)(size))

#ifdef	G2T_RGB
static Cell	***DecodeImage(Nx,Ny,interlace,codeSize,R,G,B,gray)
int		Nx,Ny,interlace,codeSize,*gray;
CM		*R,*G,*B;
#else
static Cell	**DecodeImage(Nx,Ny,interlace,codeSize)
int		Nx,Ny,interlace,codeSize;
#endif
{
	char	*cmae="cell memory allocation error.";
	int	z1,z2,z3,z,y,x,new,base,last,size,
		Bit=codeSize+1,Clr=1<<(Bit-1),End=Clr+1,Top=End+1,Wid=4097-Top,
		rem=0,num=0,len=Bit,cnt=0,old=Clr,code=Top;
	uChar	*ptr,byte[255],ring[4097-6],Tail[4096-6],*tail=Tail-6;
	long	buf=0,mask=(1<<Bit)-1;
	uShort	Head[4096-6],*head=Head-6;
#ifdef	G2T_RGB
	Cell	***cell,*r,*g,*b;
	uChar	pv;

	*gray=1;

	if ((cell=ALLOC(Cell **,3))==NULL ||
	    (cell[0]=ALLOC(Cell *,Ny))==NULL ||
	    (cell[1]=ALLOC(Cell *,Ny))==NULL ||
	    (cell[2]=ALLOC(Cell *,Ny))==NULL) Error(cmae);
#else
	Cell	**cell,*pv;

	if ((cell=ALLOC(Cell *,Ny))==NULL) Error(cmae);
#endif
	if (interlace) {
	    z1=(Ny-1)/8+1; z2=z1+(Ny-5)/8+1; z3=z2+(Ny-3)/4+1;
	}
	for (z=0; z<Ny; z++) {
	    y=(interlace)?(z<z1)?   z    *8:
			  (z<z2)?4+(z-z1)*8:
			  (z<z3)?2+(z-z2)*4:
				 1+(z-z3)*2:z;
#ifdef	G2T_RGB
	    if ((cell[0][y]=r=ALLOC(Cell,Nx))==NULL ||
		(cell[1][y]=g=ALLOC(Cell,Nx))==NULL ||
		(cell[2][y]=b=ALLOC(Cell,Nx))==NULL) Error(cmae);
#else
	    if ((cell[y]=pv=ALLOC(Cell,Nx))==NULL) Error(cmae);
#endif
	    for (x=0; x<Nx; x++) {
		while (rem==0) {
		    while (num<len) {
			if (cnt==0 && (cnt=ReadBlock(ptr=byte))==0)
			    Error("missing data block.");

			--cnt; buf|=((long)*(ptr++)<<num); num+=8;
		    }
		    num-=len; new=(int)(buf&mask); buf>>=len; 

		    if (new==End) Error("unexpected end of code.");

		    if (new==Clr) {
			len=Bit; mask=(1<<Bit)-1;
		    }
		    else if (old==Clr) {
			if (new>Clr) Error("bad code after clear code.");

			ring[base=0]=(uChar)(last=new); size=rem=1; code=Top;
		    }
		    else {
			if (code==4096) Error("missing clear code.");

			if (new>code) Error("bad code.");

			if (new==code) {
			    ring[base=(base+Wid-1)%Wid]=(uChar)last; ++size;
			}
			else {
			    base=size=0;
			    for (last=new; last>=Top; last=head[last])
				ring[size++]=tail[last];

			    ring[size++]=(uChar)last;
			}
			rem=size;

			head[code]=(uShort)old; tail[code]=(uChar)last;
			if (code==mask && len<12) {
			    ++len; mask=(mask<<1)|1;
			}
			++code;
		    }
		    old=new;
		}
#ifdef	G2T_RGB
		r[x]=R[pv=ring[(base+(--rem))%Wid]];
		g[x]=G[pv]; if (*gray && R[pv]!=G[pv]) *gray=0;
		b[x]=B[pv]; if (*gray && R[pv]!=B[pv]) *gray=0;
#else
		pv[x]=ring[(base+(--rem))%Wid];
#endif
	    }
	}
	if (rem>0)
	    (void)fputs("data remains (warning).\n",stderr);
	else {
	    while (num<len) {
		if (cnt==0 && (cnt=ReadBlock(ptr=byte))==0) {
		    cnt=255; break;
		}
		--cnt; buf|=((long)*(ptr++)<<num); num+=8;
	    }
	    if (cnt!=0 && cnt!=255)
		(void)fputs("too many codes (warning).\n",stderr);
	    else
		if (!(num>=len && (buf&mask)==End) &&
		    !(num>=len-1 && (buf&(mask>>1))==End))
		    (void)fputs("missing end of code (warning).\n",stderr);
	}
	if (cnt!=255) while (ReadBlock(byte)!=0) ;

	return cell;
}

#define PATH_LEN	1023

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	byte,rgbSize,z,Nx,Ny,codeSize,y,idx=0;
	CM	gR[256],gG[256],gB[256],lR[256],lG[256],lB[256],*R,*G,*B;
	char	path[PATH_LEN];
#ifdef	G2T_RGB
	Cell	***cell;
	int	gray;
#else
	int	BPS,bps=8;
	Cell	**cell;
#endif
	if (argc!=3) Error("usage : g2t_split GIF file_name_format");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (Get()!='G' || Get()!='I' || Get()!='F' ||
	    Get()!='8' || ((byte=Get())!='7' && byte!='9') || Get()!='a')
	    Error("bad magic number.");

	Skip(4);	/* logical screen width & height */

	byte=Get(); rgbSize=((byte>>4)&7)+1;

	Skip(2);	/* background color index & pixel aspect ratio */

	if (byte&128)
#ifdef	G2T_RGB
	    ReadCM(rgbSize,(byte&7)+1,gR,gG,gB);
	else
	    for (z=0; z<256; z++) gR[z]=gG[z]=gB[z]=(CM)z;
#else
	    bps=ReadCM(rgbSize,(byte&7)+1,gR,gG,gB);
	else
	    for (z=0; z<256; z++) gR[z]=gG[z]=gB[z]=(CM)((z<<8)|z);
#endif
	while ((byte=Get())!=';')
	    if (byte==',') {
		Skip(4);	/* image left & top position */

		Nx=Get(); Nx|=(Get()<<8);
		Ny=Get(); Ny|=(Get()<<8);

		if ((byte=Get())&128)
#ifdef	G2T_RGB
		    ReadCM(rgbSize,(byte&7)+1,R=lR,G=lG,B=lB);
		else {
		    R=gR; G=gG; B=gB;
#else
		    BPS=ReadCM(rgbSize,(byte&7)+1,R=lR,G=lG,B=lB);
		else {
		    BPS=bps; R=gR; G=gG; B=gB;
#endif
		}
		if ((codeSize=Get())<2 || codeSize>8) Error("bad code size.");
#ifdef	G2T_RGB
		cell=DecodeImage(Nx,Ny,byte&64,codeSize,R,G,B,&gray);
#else
		cell=DecodeImage(Nx,Ny,byte&64,codeSize);
#endif
		(void)sprintf(path,argv[2],idx++);
#ifdef	G2T_RGB
		if (gray)
		    StoreImageFile(path,Nx,Ny,8,*cell,NULL);
		else
		    StoreImageFile_RGB(path,Nx,Ny,8,cell[0],
						    cell[1],
						    cell[2],NULL);

		for (y=0; y<Ny; y++) {
		    free(cell[0][y]); free(cell[1][y]); free(cell[2][y]);
		}
		free(cell[0]); free(cell[1]); free(cell[2]); free(cell);
#else
		StoreImageFile_CM(path,Nx,Ny,BPS,R,G,B,cell,NULL);

		for (y=0; y<Ny; y++) free(cell[y]); free(cell);
#endif
	    }
	    else {
		if (byte=='!') Skip(1);	/* control label */

		while ((byte=Get())!=0) Skip(byte);
	    }

	if (file!=stdin) (void)fclose(file); return 0;
}
