
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ph.h"

typedef	struct Ovoid {
		int		level;
		double		xO,yO,zO,ax,ay,az,bx,by,bz,cx,cy,cz;
		unsigned short	flag;
		struct Ovoid	*next;
	} Ovoid;

static Ovoid	*Parameter(level,str,flag)
int		level,flag;
char		**str;
{
	double	A,B,C,lon,lat,rot,slon,clon,slat,clat,srot,crot;
	Ovoid	*ovoid;

	if ((A=atof(str[6]))<=0.0 ||
	    (B=atof(str[7]))<=0.0 ||
	    (C=atof(str[8]))<=0.0) {
	    fputs("bad axial radius.\n",stderr); exit(1);
	}
	if ((ovoid=(Ovoid *)malloc(sizeof(*ovoid)))==NULL) {
	    fputs("memory allocation error.\n",stderr); exit(1);
	}
	ovoid->level=level;
	ovoid->xO=atof(str[0]);
	ovoid->yO=atof(str[1]);
	ovoid->zO=atof(str[2]);
	lon=atof(str[3])*(M_PI/180.0); slon=sin(lon); clon=cos(lon);
	lat=atof(str[4])*(M_PI/180.0); slat=sin(lat); clat=cos(lat);
	rot=atof(str[5])*(M_PI/180.0); srot=sin(rot); crot=cos(rot);
	ovoid->ax=( slon*crot+clon*slat*srot)/A;
	ovoid->ay=(-clon*crot+slon*slat*srot)/A;
	ovoid->az=		  -clat*srot /A;
	ovoid->bx=(-slon*srot+clon*slat*crot)/B;
	ovoid->by=( clon*srot+slon*slat*crot)/B;
	ovoid->bz=		  -clat*crot /B;
	ovoid->cx=	      clon*clat      /C;
	ovoid->cy=	      slon*clat      /C;
	ovoid->cz=		   slat	     /C;
	ovoid->flag=(unsigned short)flag;
	ovoid->next=NULL;
	return ovoid;
}

static int	Level(str)
char		*str;
{
	int	level;

	if ((level=atoi(str))<0) {
	    fputs("bad recursion level.\n",stderr); exit(1);
	}
	return level;
}

static int	Flag(str)
char		**str;
{
	int	r,g,b;

	if ((r=atoi(str[0]))<0 || r>255 ||
	    (g=atoi(str[1]))<0 || g>255 ||
	    (b=atoi(str[2]))<0 || b>255) {
	    fputs("bad surface color.\n",stderr); exit(1);
	}
	return 32768|((r&248)<<7)|((g&248)<<2)|((b&248)>>3);
}

static Ovoid	*ovoid=NULL;

static void	Deform(xD,yD,zD,xF,yF,zF)
double		*xD,*yD,*zD;
float		*xF,*yF,*zF;
{
	double	a=ovoid->ax**xD+ovoid->ay**yD+ovoid->az**zD,
		b=ovoid->bx**xD+ovoid->by**yD+ovoid->bz**zD,
		c=ovoid->cx**xD+ovoid->cy**yD+ovoid->cz**zD,R=sqrt(a*a+b*b+c*c);

	*xF=(float)(*xD=ovoid->xO+*xD/R);
	*yF=(float)(*yD=ovoid->yO+*yD/R);
	*zF=(float)(*zD=ovoid->zO+*zD/R);
}

static void	Normalize(x,y,z,xP,yP,zP)
double		x,y,z,*xP,*yP,*zP;
{
	double	R=sqrt(x*x+y*y+z*z);

	*xP=x/R; *yP=y/R; *zP=z/R;
}

static size_t	fWriteReverse(ptr,size,items,file)
char		*ptr;
size_t		size,items;
FILE		*file;
{
	char	*p,*h,*t,c;
	size_t	i,s=size-1;

	for (p=ptr, i=0; i<items; i++, p+=size)
	    for (h=p, t=p+s; h<t; h++, t--) {
		c=(*h); *h=(*t); *t=c;
	    }

	return fwrite(ptr,size,items,file);
}

static FILE	*file;
static size_t	(*fWrite)();

static void	Draw(level,x0,y0,z0,x1,y1,z1,x2,y2,z2)
int		level;
double		x0,y0,z0,x1,y1,z1,x2,y2,z2;
{
	if (level==0)
{
	float		xyz[12];
	unsigned short	flag=ovoid->flag;

	Deform(&x0,&y0,&z0,xyz+3,xyz+ 4,xyz+ 5);
	Deform(&x1,&y1,&z1,xyz+6,xyz+ 7,xyz+ 8); x1-=x0; y1-=y0; z1-=z0;
	Deform(&x2,&y2,&z2,xyz+9,xyz+10,xyz+11); x2-=x0; y2-=y0; z2-=z0;
	Normalize(y1*z2-z1*y2,
		  z1*x2-x1*z2,
		  x1*y2-y1*x2,&x0,&y0,&z0); xyz[0]=(float)x0;
					    xyz[1]=(float)y0;
					    xyz[2]=(float)z0;
	(void)(*fWrite)((char *)xyz,sizeof(*xyz),12,file);
	(void)(*fWrite)((char *)&flag,sizeof(flag),1,file);
}
	else
{
	double	x3,y3,z3,x4,y4,z4,x5,y5,z5;

	Normalize((x0+x1)/2.0,(y0+y1)/2.0,(z0+z1)/2.0,&x3,&y3,&z3);
	Normalize((x1+x2)/2.0,(y1+y2)/2.0,(z1+z2)/2.0,&x4,&y4,&z4);
	Normalize((x2+x0)/2.0,(y2+y0)/2.0,(z2+z0)/2.0,&x5,&y5,&z5);
	Draw(--level,x3,y3,z3,x4,y4,z4,x5,y5,z5);
	Draw(  level,x0,y0,z0,x3,y3,z3,x5,y5,z5);
	Draw(  level,x1,y1,z1,x4,y4,z4,x3,y3,z3);
	Draw(  level,x2,y2,z2,x5,y5,z5,x4,y4,z4);
}}

#define STR_LEN	80

int	main(argc,argv)
int	argc;
char	**argv;
{
	Trigon	*trigon;
	int	level=0,flag=0,facets=0,trigons=Polyhedron(&trigon);

	if (argc!= 2 && argc!= 3 && argc!= 5 && argc!= 6 &&
	    argc!=11 && argc!=12 && argc!=14 && argc!=15)
{
	char	*run,*ptr;

	for (run=ptr=argv[0]; *ptr!='\0'; ptr++) if (*ptr=='/' ||
						     *ptr=='\\') run=ptr+1;

	fprintf(stderr,
"usage : %s {level} {xO yO zO lon lat rot A B C} {scR scG scB} STL_file\n",
		run);
	exit(1);
}
	if (argc==3 || argc==6 || argc==12 || argc==15) level=Level(argv[1]);

	if (argc==5 || argc==6 || argc==14 || argc==15) flag=Flag(argv+argc-4);

	if (argc==11 || argc==12 || argc==14 || argc==15)
{
	switch(argc) {
	    case 11: case 14: ovoid=Parameter(level,argv+1,flag); break;
	    case 12: case 15: ovoid=Parameter(level,argv+2,flag); break;
	}
	facets=trigons<<(level*2);
}
	else
{
	char	line[STR_LEN*13];
	Ovoid	**o=(&ovoid);

	while (gets(line)!=NULL)
{
	char	*str[13],s[13][STR_LEN];

	switch(sscanf(line,"%s %s %s %s %s %s %s %s %s %s %s %s %s",
		      str[0]=s[0],str[ 1]=s[ 1],str[ 2]=s[ 2],str[ 3]=s[ 3],
				  str[ 4]=s[ 4],str[ 5]=s[ 5],str[ 6]=s[ 6],
				  str[ 7]=s[ 7],str[ 8]=s[ 8],str[ 9]=s[ 9],
				  str[10]=s[10],str[11]=s[11],str[12]=s[12])) {
	    case  9: *o=Parameter(level,str,flag); break;
	    case 10: *o=Parameter(Level(str[0]),str+1,flag); break;
	    case 12: *o=Parameter(level,str,Flag(str+9)); break;
	    case 13: *o=Parameter(Level(str[0]),str+1,Flag(str+10)); break;
	    default: continue;
	}
	facets+=(trigons<<((*o)->level*2)); o=(&((*o)->next));
}
	if (!facets) {
	    fputs("no parameter.\n",stderr); exit(1);
	}
}
	if (argv[argc-1][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[argc-1],"wb"))==NULL) {
		fprintf(stderr,"%s : STL file not stored.\n",argv[argc-1]);
		exit(1);
	    }
{
	int	idx;

	for (idx=0; idx<80; idx++) (void)fputc(0,file);
}{
	short	word=0x1234;
	char	*byte=(char *)&word;

	fWrite=(*byte==(word&0xff))?fwrite:fWriteReverse;
}
	(void)(*fWrite)(&facets,sizeof(facets),1,file);

	do
{
	Ovoid	*next=ovoid->next;
	Trigon	*t=trigon+trigons-1;

	do Draw(ovoid->level,t->a->x,t->a->y,t->a->z,
			     t->b->x,t->b->y,t->b->z,
			     t->c->x,t->c->y,t->c->z); while (--t>=trigon);

	free(ovoid); ovoid=next;
}
	while (ovoid!=NULL);

	if (file!=stdout) fclose(file);

	return 0;
}
