
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

static size_t	fWriteReverse(ptr,size,items,file)
char		*ptr;
size_t		size,items;
FILE		*file;
{
	char	*p,*h,*t,c;
	size_t	i,s=size-1;

	for (p=ptr, i=0; i<items; i++, p+=size)
	    for (h=p, t=p+s; h<t; h++, t--) {
		c=(*h); *h=(*t); *t=c;
	    }

	return fwrite(ptr,size,items,file);
}

static FILE	*file;
static size_t	(*fWrite)();
static int	X,Y,Z,facets=0;

typedef unsigned short	Word;

static void	PrintFacet(dx,dy,dz,x0,y0,z0,x1,y1,z1,x2,y2,z2,flag)
int		dx,dy,dz,x0,y0,z0,x1,y1,z1,x2,y2,z2;
Word		flag;
{
	float	xyz[12];

	xyz[0]=(float)   dx ; xyz[ 1]=(float)   dy ; xyz[ 2]=(float)   dz ;
	xyz[3]=(float)(X+x0); xyz[ 4]=(float)(Y+y0); xyz[ 5]=(float)(Z+z0);
	xyz[6]=(float)(X+x1); xyz[ 7]=(float)(Y+y1); xyz[ 8]=(float)(Z+z1);
	xyz[9]=(float)(X+x2); xyz[10]=(float)(Y+y2); xyz[11]=(float)(Z+z2);

	if ((*fWrite)((char *)xyz,sizeof(*xyz),12,file)!=12 ||
	    (*fWrite)((char *)&flag,sizeof(flag),1,file)!=1)
	    Error("STL data write error.");

	++facets;
}

static void	PrintMinusZ(x0,y0,x1,y1,x2,y2,flag)
int		x0,y0,x1,y1,x2,y2;
Word		flag;
{
	PrintFacet(0,0,-1,x2,y2,0,x1,y1,0,x0,y0,0,flag);
}

static void	PrintPlusZ(x0,y0,x1,y1,x2,y2,flag)
int		x0,y0,x1,y1,x2,y2;
Word		flag;
{
	PrintFacet(0,0, 1,x0,y0,1,x1,y1,1,x2,y2,1,flag);
}

static void	PrintMinusX(y0,z0,y1,z1,y2,z2,flag)
int		y0,z0,y1,z1,y2,z2;
Word		flag;
{
	PrintFacet(-1,0,0,0,y2,z2,0,y1,z1,0,y0,z0,flag);
}

static void	PrintPlusX(y0,z0,y1,z1,y2,z2,flag)
int		y0,z0,y1,z1,y2,z2;
Word		flag;
{
	PrintFacet( 1,0,0,1,y0,z0,1,y1,z1,1,y2,z2,flag);
}

static void	PrintMinusY(z0,x0,z1,x1,z2,x2,flag)
int		z0,x0,z1,x1,z2,x2;
Word		flag;
{
	PrintFacet(0,-1,0,x2,0,z2,x1,0,z1,x0,0,z0,flag);
}

static void	PrintPlusY(z0,x0,z1,x1,z2,x2,flag)
int		z0,x0,z1,x1,z2,x2;
Word		flag;
{
	PrintFacet(0, 1,0,x0,1,z0,x1,1,z1,x2,1,z2,flag);
}

typedef struct List {
		int		x,y;
		struct List	*next;
	} List;

static List	*Alloc(x,y,next)
int		x,y;
List		*next;
{
	List	*list=(List *)malloc(sizeof(*list));

	if (list==NULL) Error("list memory allocation error.");

	list->x=x; list->y=y; list->next=next; return list;
}

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size_t)(size))

static void	(*Print)();

#define PRINT(a,b,c,flag)	\
	(*Print)(a->x/2,a->y/2,b->x/2,b->y/2,c->x/2,c->y/2,flag)

static void	Flush(list,flag)
List		*list;
Word		flag;
{
	int	idx,oSide,nSide,x0,y0,x1,y1,x2,y2,cnt=0;
	List	**array,**head,**tail,*oList,*nList,*stack=list;

	while ((list=list->next)!=NULL) ++cnt;

	if ((array=ALLOC(List *,cnt))==NULL)
	    Error("array memory allocation error.");

	list=stack;
	for (idx=0; idx<cnt; idx++) {
	    array[idx]=list; list=list->next;
	}
	free(list);

	stack->next=NULL;

	head=array+1; tail=array+cnt-1;
	if ((*head)->y>(*tail)->y) {
	    oSide=0; oList=(*(head++));
	}
	else {
	    oSide=1; oList=(*(tail--));
	}
	while (head!=tail) {
	    if ((*head)->y>(*tail)->y) {
		nSide=0; nList=(*(head++));
	    }
	    else {
		nSide=1; nList=(*(tail--));
	    }
	    if (oSide==nSide) {
		x2=oList->x-(x0=nList->x); y2=oList->y-(y0=nList->y);
		if (oSide==0)
		    do {
			x1=x2; y1=y2;
			if (y1*(x2=stack->x-x0)<=x1*(y2=stack->y-y0)) break;

			PRINT(stack,oList,nList,flag); oList=stack;
		    } while ((stack=stack->next)!=NULL);
		else /* if (oSide==1) */
		    do {
			x1=x2; y1=y2;
			if (y1*(x2=stack->x-x0)>=x1*(y2=stack->y-y0)) break;

			PRINT(stack,nList,oList,flag); oList=stack;
		    } while ((stack=stack->next)!=NULL);
	    }
	    else {
		list=oList;
		if (oSide==0)
		    do {
			PRINT(stack,oList,nList,flag); oList=stack;
		    } while ((stack=stack->next)!=NULL);
		else /* if (oSide==1) */
		    do {
			PRINT(stack,nList,oList,flag); oList=stack;
		    } while ((stack=stack->next)!=NULL);

		oList=list;
	    }
	    oList->next=stack; stack=oList; oList=nList; oSide=nSide;
	}
	nList=(*head);
	if (oSide==0)
	    do {
		PRINT(stack,oList,nList,flag); oList=stack;
	    } while ((stack=stack->next)!=NULL);
	else /* if (oSide==1) */
	    do {
		PRINT(stack,nList,oList,flag); oList=stack;
	    } while ((stack=stack->next)!=NULL);

	for (idx=0; idx<cnt; idx++) free(array[idx]); free(array);
}

static List	*hori,**vert,**help,**ptr;

static void	Corner(x,y,a,b,c,d,h,v)
int		x,y,h,v;
Word		a,b,c,d;
{
	List	*temp,*next;

#define VERT	vert[x]
#define HELP	help[x]

switch(((a)?8:0)|((b)?4:0)|((c)?2:0)|((d)?1:0)) {
    case 0x1:
	VERT=Alloc(x,y+1,Alloc(x,y,hori=Alloc(x+1,y,NULL)));
	ptr=(&HELP);
	break;
    case 0x2:
	VERT->y=y;
	next=HELP->next;
	if (HELP->x==next->x && HELP->y==next->y) {
	    hori=Alloc(x+1,y,Alloc(x,y,next));
	    HELP->next=Alloc(x,y,NULL); Flush(VERT,c);
	}
	else
	    hori=Alloc(x+1,y,VERT);
	break;
    case 0x3:
	if (v) {
	    VERT->y=y; VERT=Alloc(x,y+1,VERT);
	}
	ptr=(&HELP);
	break;
    case 0x4:
	hori->x=x; hori->next=VERT=Alloc(x,y+1,NULL);
	*ptr=hori;
	break;
    case 0x5:
	if (h) {
	    hori->x=x; hori=hori->next=Alloc(x+1,y,NULL);
	}
	break;
    case 0x7:
	hori->x=x; VERT->y=y;
	next=HELP->next;
	if (HELP->x==next->x && HELP->y==next->y) {
	    hori->next=Alloc(x,y,next);
	    HELP->next=Alloc(x,y,NULL); Flush(VERT,d);
	}
	else
	    hori->next=VERT;
	*ptr=hori;
	break;
    case 0x8:
	VERT->y=y; hori->x=x; Flush(hori,a);
	break;
    case 0xa:
	if (h) {
	    hori->x=x; hori=Alloc(x+1,y,hori);
	}
	break;
    case 0xb:
	hori->x=x; VERT=Alloc(x,y+1,hori);
	ptr=(&HELP); HELP=hori;
	break;
    case 0xc:
	if (v) {
	    VERT->y=y; VERT=VERT->next=Alloc(x,y+1,NULL);
	}
	break;
    case 0xd:
	VERT->y=y;
	temp=(*ptr); next=temp->next;
	if (temp->x==next->x && temp->y==next->y) {
	    Flush(Alloc(x,y,next),b);
	    temp->next=Alloc(x,y,hori=Alloc(x+1,y,NULL));
	}
	else
	    VERT->next=hori=Alloc(x+1,y,NULL);
	break;
    case 0xe:
	temp=(*ptr); next=temp->next;
	if (temp->x==next->x && temp->y==next->y)
	    hori=Alloc(x+1,y,Alloc(x,y,next));
	else
	    hori=Alloc(x+1,y,Alloc(x,y,Alloc(temp->x,temp->y,next)));
	*ptr=temp->next=Alloc(x,y,VERT=Alloc(x,y+1,NULL));
	break;
}}

static Word	**lo,**up,*AC,*BD,*ac,*bd;
static long	faces=0L;

static void	Erosion(Nx,Ny)
int		Nx,Ny;
{
	int	x,xx,y,yy,hv,loup;
	Word	*loAC,*loBD,*upAC,*upBD,A,B,C,D;

	for (xx=x=0; x<Nx; x++, xx+=2) BD[x]=bd[xx]=bd[xx+1]=0;

	for (yy=y=0; y<Ny; y++, yy+=2) {
	    loAC=lo[y-1]; loBD=lo[y];
	    upAC=up[y-1]; upBD=up[y];

	    for (xx=x=0; x<Nx; x++, xx+=2) {
		AC[x]=BD[x];
		if (loBD[x] && !upBD[x]) {
		    BD[x]=loBD[x]; ++faces;
		}
		else
		    BD[x]=0;

		A=AC[x-1]; B=BD[x-1]; C=AC[x]; D=BD[x];

		if (D && D==C) {
		    if (D==A && D==B)
			ac[xx]=D;
		    else
			ac[xx]=0;

		    ac[xx+1]=D;
		}
		else
		    ac[xx]=ac[xx+1]=0;

		hv=D!=B && D!=C;
		B=loBD[x-1]; C=loAC[x]; D=loBD[x]; loup=D && upBD[x];
		Corner(xx,yy,bd[xx-1],ac[xx-1],bd[xx],ac[xx],
		       hv || (loup && !B) || (!D && B && upBD[x-1]),
		       hv || (loup && !C) || (!D && C && upAC[x]));

		Corner(xx+1,yy,bd[xx],ac[xx],bd[xx+1],ac[xx+1],0,0);
	    }
	    Corner(xx,yy,bd[xx-1],ac[xx-1],0,0,0,0);

	    for (xx=x=0; x<Nx; x++, xx+=2) {
		A=AC[x-1]; B=BD[x-1]; C=AC[x]; D=BD[x];

		if (D) {
		    if (D==B)
			bd[xx]=D;
		    else
			bd[xx]=0;

		    bd[xx+1]=D;
		}
		else
		    bd[xx]=bd[xx+1]=0;

		Corner(xx,yy+1,ac[xx-1],bd[xx-1],ac[xx],bd[xx],0,0);

		hv=A!=B && A!=C;
		C=loAC[x]; B=loBD[x-1]; A=loAC[x-1]; loup=A && upAC[x-1];
		Corner(xx+1,yy+1,ac[xx],bd[xx],ac[xx+1],bd[xx+1],
		       hv || (loup && !C) || (!A && C && upAC[x]),
		       hv || (loup && !B) || (!A && B && upBD[x-1]));
            }
	    Corner(xx,yy+1,ac[xx-1],bd[xx-1],0,0,0,0);
	}
	for (xx=x=0; x<Nx; x++, xx+=2) {
	    Corner(xx,yy,bd[xx-1],0,bd[xx],0,0,0);
	    Corner(xx+1,yy,bd[xx],0,bd[xx+1],0,0,0);
	}
        Corner(xx,yy,bd[xx-1],0,0,0,0,0);
}

#define STR_LEN	512
#define PV_SIZE	(1<<(8*sizeof(Cell)))

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nx,Ny,Nz,pv,pv1,pv2,R1,G1,B1,R2,G2,B2,
		x1,y1,z1,x2,y2,z2,x,y,z,nx,ny,nz,N,n,x0,y0,z0;
	char	**path,line[STR_LEN],
		*imae="image memory allocation error.",
		*wmae="work memory allocation error.";
	Word	flag[PV_SIZE],**tmp;
	double	d0,d1,d2;
	Cell	***image,**cell;
	long	pels=0L;

	if (argc!=4) Error("usage : si_stl_C directory name STL");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	fprintf(stderr,"%d\t%d\t%d\n",Nx,Ny,Nz);

	flag[0]=0; for (pv=1; pv<PV_SIZE; pv++) flag[pv]=65535;

	while (gets(line)!=NULL) {
	    switch(sscanf(line,"%d %d %d %d %d %d %d %d",
			       &pv1,&pv2,&R1,&G1,&B1,&R2,&G2,&B2)) {
		case  1: pv=0; pv2=pv1; break;
		case  2: pv=0; break;
		case  4: pv=1; B2=B1=G1; G2=G1=R1; R2=R1=pv2; pv2=pv1; break;
		case  5: pv=1; B2=B1; G2=G1; R2=R1; break;
		case  8: pv=1; break;
		default: continue;
	    }
	    if (pv1<0 || pv2>=PV_SIZE || pv1>pv2) continue;

	    if (!pv)
		for (pv=pv1; pv<=pv2; pv++) flag[pv]=0;
	    else {
		if (R1<0 || R1>255 || R2<0 || R2>255 ||
		    G1<0 || G1>255 || G2<0 || G2>255 ||
		    B1<0 || B1>255 || B2<0 || B2>255) continue;

#define FLAG(r,g,b)	(Word)(32768|((int)((r)*31.0/255.0+0.5)<<10)|\
				     ((int)((g)*31.0/255.0+0.5)<< 5)|\
				      (int)((b)*31.0/255.0+0.5)      )

		if (pv1==pv2)
		    flag[pv1]=FLAG((double)R1,(double)G1,(double)B1);
		else {
		    d0=(double)(pv2-pv1);
		    for (pv=pv1; pv<=pv2; pv++) {
			d1=(double)(pv2-pv)/d0;
			d2=(double)(pv-pv1)/d0;
			flag[pv]=FLAG((double)R1*d1+(double)R2*d2,
				      (double)G1*d1+(double)G2*d2,
				      (double)B1*d1+(double)B2*d2);
		    }
		}
	    }
	}

	if ((image=ALLOC(Cell **,Nz))==NULL) Error(imae);

	x1=Nx-1;
	y1=Ny-1;
	z1=Nz-1; x2=y2=z2=0;
	for (z=0; z<Nz; z++) {
	    ReadImageFile(path[z],&nx,&ny,NULL,&cell,NULL);

	    if (nx==Nx && ny==Ny)
		image[z]=cell;
	    else {
		if ((image[z]=ALLOC(Cell *,Ny))==NULL) Error(imae);

		for (y=0; y<Ny; y++) {
		    if ((image[z][y]=ALLOC(Cell,Nx))==NULL) Error(imae);

		    for (x=0; x<Nx; x++)
			image[z][y][x]=(x<nx && y<ny)?cell[y][x]:0;

		    free(cell[y]);
		}
		free(cell);
	    }
	    for (y=0; y<Ny; y++)
		for (x=0; x<Nx; x++)
		    if (flag[image[z][y][x]]) {
			if (x<x1) x1=x; if (x>x2) x2=x;
			if (y<y1) y1=y; if (y>y2) y2=y;
			if (z<z1) z1=z; if (z>z2) z2=z; ++pels;
		    }
	}
	fprintf(stderr,"%ld\t%d\t%d\t%d\t%d\t%d\t%d\n",pels,x1,y1,z1,x2,y2,z2);

	if (pels==0L) Error("no object exists.");

	nx=x2-x1+1;
	ny=y2-y1+1;
	nz=z2-z1+1; N=(nx>ny)?(nx>nz)?nx:nz:(ny>nz)?ny:nz;

	if ((lo=ALLOC(Word *,N+1))==NULL ||
	    (up=ALLOC(Word *,N+1))==NULL ||
	    (AC=ALLOC(Word,N+1))==NULL ||
	    (BD=ALLOC(Word,N+1))==NULL ||
	    (ac=ALLOC(Word,N*2+1))==NULL ||
	    (bd=ALLOC(Word,N*2+1))==NULL ||
	    (vert=ALLOC(List *,N*2+1))==NULL ||
	    (help=ALLOC(List *,N*2))==NULL) Error(wmae);

	++lo; ++up;

	for (n=(-1); n<N; n++) {
	    if ((lo[n]=ALLOC(Word,N+1))==NULL ||
		(up[n]=ALLOC(Word,N+1))==NULL) Error(wmae);

	    ++lo[n]; ++up[n]; lo[-1][n]=lo[n][-1]=up[-1][n]=up[n][-1]=0;
	}
	++AC; ++BD; ++ac; ++bd; AC[-1]=BD[-1]=ac[-1]=bd[-1]=0;

	if ((file=fopen(argv[3],"wb"))==NULL) Error("STL file not stored.");
{
	int	i;
	short	s=0x1234;
	char	*c=(char *)&s;

	for (i=80+sizeof(facets); i>0; i--)
	    if (fputc(0,file)!=0) Error("STL header data write error.");

	fWrite=(*c==(s&0xff))?fwrite:fWriteReverse;
}

X=x1; Y=y1;

	Print=PrintMinusZ;

	for (y=0; y<ny; y++)
	for (x=0; x<nx; x++) lo[y][x]=0;

	for (Z=z1; Z<=z2; Z++) {
	    tmp=up; up=lo; lo=tmp;

	    for (y0=Y, y=0; y<ny; y++, y0++)
	    for (x0=X, x=0; x<nx; x++, x0++) lo[y][x]=flag[image[Z][y0][x0]];

	    Erosion(nx,ny);
	}

	Print=PrintPlusZ;

	for (y=0; y<ny; y++)
	for (x=0; x<nx; x++) lo[y][x]=0;

	for (Z=z2; Z>=z1; Z--) {
	    tmp=up; up=lo; lo=tmp;

	    for (y0=Y, y=0; y<ny; y++, y0++)
	    for (x0=X ,x=0; x<nx; x++, x0++) lo[y][x]=flag[image[Z][y0][x0]];

	    Erosion(nx,ny);
	}

Y=y1; Z=z1;

	Print=PrintMinusX;

	for (z=0; z<nz; z++)
	for (y=0; y<ny; y++) lo[z][y]=0;

	for (X=x1; X<=x2; X++) {
	    tmp=up; up=lo; lo=tmp;

	    for (z0=Z, z=0; z<nz; z++, z0++)
	    for (y0=Y, y=0; y<ny; y++, y0++) lo[z][y]=flag[image[z0][y0][X]];

	    Erosion(ny,nz);
	}

	Print=PrintPlusX;

	for (z=0; z<nz; z++)
	for (y=0; y<ny; y++) lo[z][y]=0;

	for (X=x2; X>=x1; X--) {
	    tmp=up; up=lo; lo=tmp;

	    for (z0=Z, z=0; z<nz; z++, z0++)
	    for (y0=Y, y=0; y<ny; y++, y0++) lo[z][y]=flag[image[z0][y0][X]];

	    Erosion(ny,nz);
	}

Z=z1; X=x1;

	Print=PrintMinusY;

	for (x=0; x<nx; x++)
	for (z=0; z<nz; z++) lo[x][z]=0;

	for (Y=y1; Y<=y2; Y++) {
	    tmp=up; up=lo; lo=tmp;

	    for (x0=X, x=0; x<nx; x++, x0++)
	    for (z0=Z, z=0; z<nz; z++, z0++) lo[x][z]=flag[image[z0][Y][x0]];

	    Erosion(nz,nx);
	}

	Print=PrintPlusY;

	for (x=0; x<nx; x++)
	for (z=0; z<nz; z++) lo[x][z]=0;

	for (Y=y2; Y>=y1; Y--) {
	    tmp=up; up=lo; lo=tmp;

	    for (x0=X, x=0; x<nx; x++, x0++)
	    for (z0=Z, z=0; z<nz; z++, z0++) lo[x][z]=flag[image[z0][Y][x0]];

	    Erosion(nz,nx);
	}

	fprintf(stderr,"%ld\t%d\n",faces,facets);

	if ((file=freopen(argv[3],"rb+",file))==NULL ||
	    fseek(file,80L,SEEK_SET) ||
	    (*fWrite)(&facets,sizeof(facets),1,file)!=1)
	    Error("STL trailer data write error.");

	fclose(file);

	return 0;
}
