
#ifndef	STL_COLOR
#define STL_COLOR
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "stl.h"
#include "xyz.h"
#include "cell.h"
#include "sif_rgb.h"

#define gamma	Gamma	/* to avoid duplication error */

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

static double	gamma,bias;

static void	RGB(shade,ratioR,ratioG,ratioB,pvR,pvG,pvB)
double		shade,ratioR,ratioG,ratioB;
Cell		*pvR,*pvG,*pvB;
{
	double	intensity=bias+(255.0-bias)*pow(shade,1.0/gamma);

	*pvR=(Cell)(intensity*ratioR+0.5);
	*pvG=(Cell)(intensity*ratioG+0.5);
	*pvB=(Cell)(intensity*ratioB+0.5);
}

static Cell	**cellR,**cellG,**cellB;
static int	Ny1;

#ifdef	SS
static void	Shade(x,y,shade)
int		x,y;
double		shade;
{
	Cell	*R=(&(cellR[Ny1-y][x])),
		*G=(&(cellG[Ny1-y][x])),
		*B=(&(cellB[Ny1-y][x]));

	RGB(shade,(double)*R/255.0,(double)*G/255.0,(double)*B/255.0,R,G,B);
}
#endif

static Cell	pvR,pvG,pvB;

static void	Plot(x,y)
int		x,y;
{
	cellR[Ny1-y][x]=pvR;
	cellG[Ny1-y][x]=pvG;
	cellB[Ny1-y][x]=pvB;
}

static double	Distance(x,y,z,D)
double		x,y,z,D;
{
	double	d=sqrt(x*x+y*y+z*z);

	return (d>D)?d:D;
}

static void	Rotate(C,x,y,z,X,Y,Z)
double		C[3][3],x,y,z,*X,*Y,*Z;
{
	*X=C[0][0]*x+C[0][1]*y+C[0][2]*z;
	*Y=C[1][0]*x+C[1][1]*y+C[1][2]*z;
	*Z=C[2][0]*x+C[2][1]*y+C[2][2]*z;
}

#define STR_LEN		256
#define RADIAN		(M_PI/180.0)
#define MA(num,var)	malloc((size_t)(num)*sizeof(*(var)))

int	main(argc,argv)
int	argc;
char	**argv;
{
	double	xO,yO,zO,Lon,lon,Lat,lat,scale,
		D,H,V,T0,dT,L,l,P,p,cl,sl,cp,sp,A[3][3],X,Y,Z,R,r,d,
		T,t,ct,st,B[2][3],C[3][3],X0,Y0,Z0,X1,Y1,Z1,X2,Y2,Z2;
	int	bgR,bgG,bgB,fgR,fgG,fgB,y,cnt,idx,Nx,Ny,x;
	FILE	*file;
	Facet	*facet,*top=NULL,**ptr=(&top);
	char	line[STR_LEN],str[6][STR_LEN],*fmt,path[STR_LEN],desc[STR_LEN],
		*cmae="cell memory allocation error.";
#ifndef	SS
	double	scR,scG,scB,shade;
#else
	int	scR,scG,scB;
#endif
	if (argc<19)
	    Error("usage : stl_ar_C STL_files xO yO zO lon lat \\\n"
		  "                           scale gamma bias bgR bgG bgB"
							     " fgR fgG fgB"
							     " scR scG scB");

	xO=atof(argv[argc-17]);
	yO=atof(argv[argc-16]);
	zO=atof(argv[argc-15]);
	lon=(Lon=atof(argv[argc-14]))*RADIAN;
	lat=(Lat=atof(argv[argc-13]))*RADIAN;

	if ((scale=atof(argv[argc-12]))<=0.0) Error("bad scale.");
	if ((gamma=atof(argv[argc-11]))<=0.0) Error("bad gamma.");
	if ((bias=atof(argv[argc-10]))<0.0 || bias>255.0) Error("bad bias.");
	if ((bgR=atoi(argv[argc-9]))<0 || bgR>255 ||
	    (bgG=atoi(argv[argc-8]))<0 || bgG>255 ||
	    (bgB=atoi(argv[argc-7]))<0 || bgB>255) Error("bad background.");
	if ((fgR=atoi(argv[argc-6]))>255 ||
	    (fgG=atoi(argv[argc-5]))>255 ||
	    (fgB=atoi(argv[argc-4]))>255) Error("bad foreground.");
#ifndef	SS
	if ((scR=atof(argv[argc-3])/255.0)<0.0 || scR>1.0 ||
	    (scG=atof(argv[argc-2])/255.0)<0.0 || scG>1.0 ||
	    (scB=atof(argv[argc-1])/255.0)<0.0 || scB>1.0)
#else
	if ((scR=atoi(argv[argc-3]))<0 || scR>255 ||
	    (scG=atoi(argv[argc-2]))<0 || scG>255 ||
	    (scB=atoi(argv[argc-1]))<0 || scB>255)
#endif
	    Error("bad surface color.");

	while (--argc>17) {
	    if ((file=fopen(*(++argv),"rb"))==NULL) {
		fprintf(stderr,"%s : file not found.\n",*argv); exit(1);
	    }
	    if ((*ptr=ReadFacet(file))==NULL) {
		fprintf(stderr,"%s : file read error.\n",*argv); exit(1);
	    }
	    fclose(file);

	    for (facet=(*ptr); facet->next!=NULL; facet=facet->next) ;

	    ptr=(&(facet->next));
	}
	D=0.0;
	for (facet=top; facet!=NULL; facet=facet->next) {
	    D=Distance(facet->x0-xO,facet->y0-yO,facet->z0-zO,D);
	    D=Distance(facet->x1-xO,facet->y1-yO,facet->z1-zO,D);
	    D=Distance(facet->x2-xO,facet->y2-yO,facet->z2-zO,D);
	}

	XYZWidth(lon,lat,-D,-D,-D,D,D,D,&H,&V);
	Nx=(int)ceil(scale*H);
	Ny=(int)ceil(scale*V); Ny1=Ny-1;

	if ((cellR=(Cell **)MA(Ny,cellR))==NULL ||
	    (cellG=(Cell **)MA(Ny,cellG))==NULL ||
	    (cellB=(Cell **)MA(Ny,cellB))==NULL) Error(cmae);

	for (y=0; y<Ny; y++)
	    if ((cellR[y]=(Cell *)MA(Nx,*cellR))==NULL ||
		(cellG[y]=(Cell *)MA(Nx,*cellG))==NULL ||
		(cellB[y]=(Cell *)MA(Nx,*cellB))==NULL) Error(cmae);

	while (gets(line)!=NULL)
{
	switch(sscanf(line,"%s %s %s %s %s %s",
			   str[0],str[1],str[2],str[3],str[4],str[5])) {
	    case 4:
		T0=atof(str[2]); dT=0.0; cnt=1; fmt=str[3];
		break;
	    case 6:
		T0=atof(str[2]); dT=atof(str[3]); cnt=atoi(str[4]); fmt=str[5];
		break;
	    default:
		continue;
	}
	l=(L=atof(str[0]))*RADIAN; cl=cos(l); sl=sin(l);
	p=(P=atof(str[1]))*RADIAN; cp=cos(p); sp=sin(p);
	A[0][0]=(-sl); A[0][1]=(-cl*sp); A[0][2]=cl*cp;
	A[1][0]=  cl ; A[1][1]=(-sl*sp); A[1][2]=sl*cp;
	A[2][0]= 0.0 ; A[2][1]=     cp ; A[2][2]=   sp;

	R=fabs(X=A[0][2]);
	if ((r=fabs(Y=A[1][2]))>R) R=r;
	if ((r=fabs(Z=A[2][2]))>R) R=r; d=D/R; X*=d; Y*=d; Z*=d;

	for (idx=0; idx<cnt; idx++) {
	    t=(T=T0+dT*(double)idx)*RADIAN; ct=cos(t); st=sin(t);
	    B[0][0]=ct*A[0][0]-st*A[0][1];
	    B[0][1]=ct*A[1][0]-st*A[1][1];
	    B[0][2]=ct*A[2][0]-st*A[2][1];
	    B[1][0]=st*A[0][0]+ct*A[0][1];
	    B[1][1]=st*A[1][0]+ct*A[1][1];
	    B[1][2]=st*A[2][0]+ct*A[2][1];
	    C[0][0]=A[0][0]*B[0][0]+A[0][1]*B[1][0]+A[0][2]*A[0][2];
	    C[0][1]=A[0][0]*B[0][1]+A[0][1]*B[1][1]+A[0][2]*A[1][2];
	    C[0][2]=A[0][0]*B[0][2]+A[0][1]*B[1][2]+A[0][2]*A[2][2];
	    C[1][0]=A[1][0]*B[0][0]+A[1][1]*B[1][0]+A[1][2]*A[0][2];
	    C[1][1]=A[1][0]*B[0][1]+A[1][1]*B[1][1]+A[1][2]*A[1][2];
	    C[1][2]=A[1][0]*B[0][2]+A[1][1]*B[1][2]+A[1][2]*A[2][2];
	    C[2][0]=A[2][0]*B[0][0]+A[2][1]*B[1][0]+A[2][2]*A[0][2];
	    C[2][1]=A[2][0]*B[0][1]+A[2][1]*B[1][1]+A[2][2]*A[1][2];
	    C[2][2]=A[2][0]*B[0][2]+A[2][1]*B[1][2]+A[2][2]*A[2][2];

	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) {
		cellR[y][x]=(Cell)bgR;
		cellG[y][x]=(Cell)bgG;
		cellB[y][x]=(Cell)bgB;
	    }
	    if (!XYZInit(Nx,Ny,lon,lat,-D,-D,-D,D,D,D))
		Error("rendering memory allocation error.");

	    for (facet=top; facet!=NULL; facet=facet->next) {
		Rotate(C,facet->x0-xO,facet->y0-yO,facet->z0-zO,&X0,&Y0,&Z0);
		Rotate(C,facet->x1-xO,facet->y1-yO,facet->z1-zO,&X1,&Y1,&Z1);
		Rotate(C,facet->x2-xO,facet->y2-yO,facet->z2-zO,&X2,&Y2,&Z2);
#ifndef	SS
		if ((shade=XYZSide(lon,lat,X0,Y0,Z0,X1,Y1,Z1,X2,Y2,Z2))>=0.0) {
		    if (STL_COLOR_FLAG(facet->flag))
			RGB(shade,STL_COLOR_R(facet->flag),
				  STL_COLOR_G(facet->flag),
				  STL_COLOR_B(facet->flag),&pvR,&pvG,&pvB);
		    else
			RGB(shade,scR,scG,scB,&pvR,&pvG,&pvB);

		    XYZFill(X0,Y0,Z0,X1,Y1,Z1,X2,Y2,Z2,Plot);
		}
	    }
#else
		if (XYZSide(lon,lat,X0,Y0,Z0,X1,Y1,Z1,X2,Y2,Z2)>=0.0) {
		    if (STL_COLOR_FLAG(facet->flag)) {
			pvR=(Cell)(STL_COLOR_R(facet->flag)*255.0+0.5);
			pvG=(Cell)(STL_COLOR_G(facet->flag)*255.0+0.5);
			pvB=(Cell)(STL_COLOR_B(facet->flag)*255.0+0.5);
		    }
		    else {
			pvR=(Cell)scR;
			pvG=(Cell)scG;
			pvB=(Cell)scB;
		    }
		    XYZFill(X0,Y0,Z0,X1,Y1,Z1,X2,Y2,Z2,Plot);
		}
	    }
	    XYZShade(lon,lat,Shade);
#endif
	    if (fgR>=0 && fgG>=0 && fgB>=0) {
		pvR=(Cell)fgR;
		pvG=(Cell)fgG;
		pvB=(Cell)fgB;
		XYZDraw(-D,-D,-D, D,-D,-D,Plot);
		XYZDraw( D,-D,-D, D, D,-D,Plot);
		XYZDraw( D, D,-D,-D, D,-D,Plot);
		XYZDraw(-D, D,-D,-D,-D,-D,Plot);
		XYZDraw(-D,-D, D, D,-D, D,Plot);
		XYZDraw( D,-D, D, D, D, D,Plot);
		XYZDraw( D, D, D,-D, D, D,Plot);
		XYZDraw(-D, D, D,-D,-D, D,Plot);
		XYZDraw(-D,-D,-D,-D,-D, D,Plot);
		XYZDraw( D,-D,-D, D,-D, D,Plot);
		XYZDraw( D, D,-D, D, D, D,Plot);
		XYZDraw(-D, D,-D,-D, D, D,Plot); XYZDraw(-X,-Y,-Z,X,Y,Z,Plot);
	    }
	    XYZTerm();

	    sprintf(path,fmt,idx);
	    sprintf(desc,
		    "%f\t%f\t%f\n%f\t%f\t%f\n%f\t%f\n",xO,yO,zO,L,P,T,Lon,Lat);
	    StoreImageFile_RGB(path,Nx,Ny,8,cellR,cellG,cellB,desc);
	}
}
	return 0;
}
