
#ifndef	STL_COLOR
#define STL_COLOR
#endif

#ifdef	LINE_GRAY
#undef	LINE_GRAY
#endif
#ifndef	LINE_RGB
#define	LINE_RGB
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "stl.h"
#include "line.h"
#include "xyz.h"
#include "sgif.h"

#define gamma	Gamma	/* to avoid duplication error */

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

typedef unsigned char	byte;

static double	gamma,bias;

static void	RGB(shade,ratioR,ratioG,ratioB,pvR,pvG,pvB)
double		shade,ratioR,ratioG,ratioB;
byte		*pvR,*pvG,*pvB;
{
	double	intensity=bias+(255.0-bias)*pow(shade,1.0/gamma);

	*pvR=(byte)(intensity*ratioR+0.5);
	*pvG=(byte)(intensity*ratioG+0.5);
	*pvB=(byte)(intensity*ratioB+0.5);
}

static byte	**cellR,**cellG,**cellB;
static int	xL,yT;

#ifdef	SS
static void	Shade(x,y,shade)
int		x,y;
double		shade;
{
	byte	*R=(&(cellR[yT-y][xL+x])),
		*G=(&(cellG[yT-y][xL+x])),
		*B=(&(cellB[yT-y][xL+x]));

	RGB(shade,(double)*R/255.0,(double)*G/255.0,(double)*B/255.0,R,G,B);
}
#endif

static byte	pvR,pvG,pvB;

static void	Plot(x,y)
int		x,y;
{
	cellR[yT-y][xL+x]=pvR;
	cellG[yT-y][xL+x]=pvG;
	cellB[yT-y][xL+x]=pvB;
}

#define STR_LEN		256
#define MA(num,var)	malloc((size_t)(num)*sizeof(*(var)))

int	main(argc,argv)
int	argc;
char	**argv;
{
	double	scale,x1,y1,z1,x2,y2,z2,dx,dy,dz,lo0,la0,dLo,dLa,lo,la,h,v;
	int	bgR,bgG,bgB,fgR,fgG,fgB,idx,no,y,cnt,Nx,Ny,x,size=0;
	FILE	*file;
	Facet	*facet,*fTop=NULL,**fPtr=(&fTop);
	Line	*line,*lTop=NULL,**lPtr=(&lTop);
	float	*lx,*ly,*lz;
	char	*name,text[STR_LEN],str[5][STR_LEN],
		*fnf="%s : file not found.\n",
		*fre="%s : file read error.\n",
		*cmae="cell memory allocation error.";
#ifndef	SS
	double	scR,scG,scB,shade;
#else
	int	scR,scG,scB;
#endif

	if (argc<15)
	    Error("usage : stl_bev_C_GIF STL_files scale gamma bias \\\n"
		  "                                bgR bgG bgB"
						 " fgR fgG fgB"
						 " scR scG scB GIF");

	if (argv[argc-13][0]!='-') {
	    if ((scale=atof(argv[argc-13]))<=0.0) Error("bad scale.");
	}
	else
	    if ((size=atoi(argv[argc-13]+1))<=0) Error("bad size.");

	if ((gamma=atof(argv[argc-12]))<=0.0) Error("bad gamma.");
	if ((bias=atof(argv[argc-11]))<0.0 || bias>255.0) Error("bad bias.");
	if ((bgR=atoi(argv[argc-10]))<0 || bgR>255 ||
	    (bgG=atoi(argv[argc- 9]))<0 || bgG>255 ||
	    (bgB=atoi(argv[argc- 8]))<0 || bgB>255) Error("bad background.");
	if ((fgR=atoi(argv[argc-7]))>255 ||
	    (fgG=atoi(argv[argc-6]))>255 ||
	    (fgB=atoi(argv[argc-5]))>255) Error("bad foreground.");
#ifndef	SS
	if ((scR=atof(argv[argc- 4])/255.0)<0.0 || scR>1.0 ||
	    (scG=atof(argv[argc- 3])/255.0)<0.0 || scG>1.0 ||
	    (scB=atof(argv[argc- 2])/255.0)<0.0 || scB>1.0)
#else
	if ((scR=atoi(argv[argc- 4]))<0 || scR>255 ||
	    (scG=atoi(argv[argc- 3]))<0 || scG>255 ||
	    (scB=atoi(argv[argc- 2]))<0 || scB>255)
#endif
	    Error("bad surface color.");

	for (idx=1; idx<argc-13; idx++)
	    if (*(name=argv[idx])!='-') {
		if ((file=fopen(name,"rb"))==NULL) {
		    fprintf(stderr,fnf,name); exit(1);
		}
		if ((*fPtr=ReadFacet(file))==NULL) {
		    fprintf(stderr,fre,name); exit(1);
		}
		fclose(file);

		for (facet=(*fPtr); facet->next!=NULL; facet=facet->next) ;

		fPtr=(&(facet->next));
	    }
	    else {
		if ((file=fopen(name+1,"r"))==NULL) {
		    fprintf(stderr,fnf,name); exit(1);
		}
		if ((*lPtr=ReadLine(file))==NULL) {
		    fprintf(stderr,fre,name); exit(1);
		}
		fclose(file);
  
		for (line=(*lPtr); line->next!=NULL; line=line->next) ;
 
		lPtr=(&(line->next));
	    }

	if (fTop==NULL) Error("no facet");

	x1=x2=fTop->x0;
	y1=y2=fTop->y0;
	z1=z2=fTop->z0;
	for (facet=fTop; facet!=NULL; facet=facet->next) {
	    if (facet->x0<x1) x1=facet->x0; if (facet->x0>x2) x2=facet->x0;
	    if (facet->y0<y1) y1=facet->y0; if (facet->y0>y2) y2=facet->y0;
	    if (facet->z0<z1) z1=facet->z0; if (facet->z0>z2) z2=facet->z0;
	    if (facet->x1<x1) x1=facet->x1; if (facet->x1>x2) x2=facet->x1;
	    if (facet->y1<y1) y1=facet->y1; if (facet->y1>y2) y2=facet->y1;
	    if (facet->z1<z1) z1=facet->z1; if (facet->z1>z2) z2=facet->z1;
	    if (facet->x2<x1) x1=facet->x2; if (facet->x2>x2) x2=facet->x2;
	    if (facet->y2<y1) y1=facet->y2; if (facet->y2>y2) y2=facet->y2;
	    if (facet->z2<z1) z1=facet->z2; if (facet->z2>z2) z2=facet->z2;
	}
	for (line=lTop; line!=NULL; line=line->next)
	    for (lx=line->x,
		 ly=line->y,
		 lz=line->z, no=line->No; no>0; no--, lx++, ly++, lz++) {
		if (*lx<x1) x1=(*lx); if (*lx>x2) x2=(*lx);
		if (*ly<y1) y1=(*ly); if (*ly>y2) y2=(*ly);
		if (*lz<z1) z1=(*lz); if (*lz>z2) z2=(*lz);
	    }

	dx=x2-x1;
	dy=y2-y1;
	dz=z2-z1;

	if (size)
	    scale=(double)size/sqrt(dx*dx+dy*dy+dz*dz);
	else
	    size=(int)ceil(scale*sqrt(dx*dx+dy*dy+dz*dz));

	if ((cellR=(byte **)MA(size,cellR))==NULL ||
	    (cellG=(byte **)MA(size,cellG))==NULL ||
	    (cellB=(byte **)MA(size,cellB))==NULL) Error(cmae);

	for (y=0; y<size; y++)
	    if ((cellR[y]=(byte *)MA(size,*cellR))==NULL ||
		(cellG[y]=(byte *)MA(size,*cellG))==NULL ||
		(cellB[y]=(byte *)MA(size,*cellB))==NULL) Error(cmae);

	if (argv[argc-1][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[argc-1],"wb"))==NULL)
		Error("GIF file not created.");

	InitStoreGIF(file,size,size,NULL,NULL,NULL);

	while (gets(text)!=NULL)
{
	switch(sscanf(text,"%s %s %s %s %s",
			   str[0],str[1],str[2],str[3],str[4])) {
	    case 2:
		lo0=atof(str[0]); dLo=0.0;
		la0=atof(str[1]); dLa=0.0; cnt=1; break;
	    case 5:
		lo0=atof(str[0]); dLo=atof(str[2]);
		la0=atof(str[1]); dLa=atof(str[3]); cnt=atoi(str[4]); break;
	    default:
		continue;
	}
	for (idx=0; idx<cnt; idx++) {
	    for (y=0; y<size; y++)
	    for (x=0; x<size; x++) {
		cellR[y][x]=(byte)bgR;
		cellG[y][x]=(byte)bgG;
		cellB[y][x]=(byte)bgB;
	    }
	    lo=(lo0+dLo*(double)idx)*M_PI/180.0;
	    la=(la0+dLa*(double)idx)*M_PI/180.0;

	    XYZWidth(lo,la,x1,y1,z1,x2,y2,z2,&h,&v);
	    Nx=(int)ceil(scale*h); xL=(size-Nx)/2;
	    Ny=(int)ceil(scale*v); yT=(size+Ny)/2-1;

	    if (!XYZInit(Nx,Ny,lo,la,x1,y1,z1,x2,y2,z2))
		Error("rendering memory allocation error.");

	    for (facet=fTop; facet!=NULL; facet=facet->next)
#ifndef	SS
		if ((shade=XYZSide(lo,la,
				   facet->x0,facet->y0,facet->z0,
				   facet->x1,facet->y1,facet->z1,
				   facet->x2,facet->y2,facet->z2))>=0.0) {
		    if (STL_COLOR_FLAG(facet->flag))
			RGB(shade,STL_COLOR_R(facet->flag),
				  STL_COLOR_G(facet->flag),
				  STL_COLOR_B(facet->flag),&pvR,&pvG,&pvB);
		    else
			RGB(shade,scR,scG,scB,&pvR,&pvG,&pvB);

		    XYZFill(facet->x0,facet->y0,facet->z0,
			    facet->x1,facet->y1,facet->z1,
			    facet->x2,facet->y2,facet->z2,Plot);
		}
#else
		if (XYZSide(lo,la,
			    facet->x0,facet->y0,facet->z0,
			    facet->x1,facet->y1,facet->z1,
			    facet->x2,facet->y2,facet->z2)>=0.0) {
		    if (STL_COLOR_FLAG(facet->flag)) {
			pvR=(byte)(STL_COLOR_R(facet->flag)*255.0+0.5);
			pvG=(byte)(STL_COLOR_G(facet->flag)*255.0+0.5);
			pvB=(byte)(STL_COLOR_B(facet->flag)*255.0+0.5);
		    }
		    else {
			pvR=(byte)scR;
			pvG=(byte)scG;
			pvB=(byte)scB;
		    }
		    XYZFill(facet->x0,facet->y0,facet->z0,
			    facet->x1,facet->y1,facet->z1,
			    facet->x2,facet->y2,facet->z2,Plot);
		}

	    XYZShade(lo,la,Shade);
#endif
	    for (line=lTop; line!=NULL; line=line->next) {
		pvR=(byte)line->r;
		pvG=(byte)line->g;
		pvB=(byte)line->b; lx=line->x; ly=line->y; lz=line->z;
		if (line->No==1)
		    XYZPlot((double)*lx,(double)*ly,(double)*lz,Plot);
		else
		    for (no=line->No; no>1; no--, lx++, ly++, lz++)
			XYZDraw((double)lx[0],(double)ly[0],(double)lz[0],
				(double)lx[1],(double)ly[1],(double)lz[1],Plot);
	    }
	    if (fgR>=0 && fgG>=0 && fgB>=0) {
		pvR=(byte)fgR;
		pvG=(byte)fgG;
		pvB=(byte)fgB;
		XYZDraw(x1,y1,z1,x2,y1,z1,Plot);
		XYZDraw(x2,y1,z1,x2,y2,z1,Plot);
		XYZDraw(x2,y2,z1,x1,y2,z1,Plot);
		XYZDraw(x1,y2,z1,x1,y1,z1,Plot);
		XYZDraw(x1,y1,z2,x2,y1,z2,Plot);
		XYZDraw(x2,y1,z2,x2,y2,z2,Plot);
		XYZDraw(x2,y2,z2,x1,y2,z2,Plot);
		XYZDraw(x1,y2,z2,x1,y1,z2,Plot);
		XYZDraw(x1,y1,z1,x1,y1,z2,Plot);
		XYZDraw(x2,y1,z1,x2,y1,z2,Plot);
		XYZDraw(x2,y2,z1,x2,y2,z2,Plot);
		XYZDraw(x1,y2,z1,x1,y2,z2,Plot);
	    }
	    XYZTerm();

	    StoreGIF_RGB(file,0,0,size,size,cellR,cellG,cellB);
	}
}
	TermStoreGIF(file);

	if (file!=stdout) fclose(file); return 0;
}
