
#ifndef	STL_COLOR
#define STL_COLOR
#endif

#include <stdio.h>
#include <stdlib.h>
#include "stl.h"

static void	Abort(path,msg)
char		*path,*msg;
{
	fprintf(stderr,"%s : %s.\n",path,msg); exit(1);
}

#define BYTE	unsigned char

#define EOL	'\n'	/* '\r' */
#define FLOAT	"float"	/* "float32" */
#define UCHAR	"uchar"	/* "uint8" */
#define INT	"int"	/* "int32" */

#define ORG	argv[1]
#define NEW	argv[argc-1]

#define NUM	stl.u.b.num

int	main(argc,argv)
int	argc;
char	**argv;
{
	FILE	*org=stdin,*new=stdout;
	STL	stl;
	int	num,cnt,r=255,g=255,b=255;
	float	v0[3],v1[3],v2[3];
	short	cf;
	BYTE	rgb[3];

	if (argc!=3 && argc!=6) {
	    fputs("usage : stl_ply STL {R G B} PLY\n",stderr); exit(1);
	}
	if (ORG[0]!='-')
	    if ((org=fopen(ORG,"rb"))==NULL) Abort(ORG,"file not found");

	if (!SetUpSTL(&stl,org)) Abort(ORG,"STL header read error");

	if (stl.type!=STL_TYPE_BINARY) Abort(ORG,"not a binary STL");

	if (argc==6)
	    if ((r=atoi(argv[2]))<0 || r>255 ||
		(g=atoi(argv[3]))<0 || g>255 ||
		(b=atoi(argv[4]))<0 || b>255) {
		fputs("bad surface color.\n",stderr); exit(1);
	    }

	if (NEW[0]!='-')
	    if ((new=fopen(NEW,"wb"))==NULL) Abort(NEW,"file not stored");

	fprintf(new,"ply%c",EOL);
{
	short	s=0x1234;
	char	*c=(char *)&s;

	fprintf(new,"format binary_%s_endian 1.0%c",
		    (*c==(char)(s&0xff))?"little":"big",EOL);
}
	fprintf(new,"element vertex %d%c",num=NUM*3,EOL);
	fprintf(new,"property " FLOAT " x%c",EOL);
	fprintf(new,"property " FLOAT " y%c",EOL);
	fprintf(new,"property " FLOAT " z%c",EOL);
	fprintf(new,"property " UCHAR " red%c",EOL);
	fprintf(new,"property " UCHAR " green%c",EOL);
	fprintf(new,"property " UCHAR " blue%c",EOL);
	fprintf(new,"element face %d%c",NUM,EOL);
	fprintf(new,"property list " UCHAR " " INT " vertex_indices%c",EOL);
	fprintf(new,"end_header%c",EOL);

	for (cnt=0; cnt<num; cnt+=3) {
	    if (ScanSTL(&stl,v0,v0+1,v0+2,v1,v1+1,v1+2,v2,v2+1,v2+2,&cf)<=0)
		Abort(ORG,"unexpected end of file");

	    if (STL_COLOR_FLAG(cf)) {
		rgb[0]=(BYTE)(STL_COLOR_R(cf)*255.0+0.5);
		rgb[1]=(BYTE)(STL_COLOR_G(cf)*255.0+0.5);
		rgb[2]=(BYTE)(STL_COLOR_B(cf)*255.0+0.5);
	    }
	    else {
		rgb[0]=(BYTE)r;
		rgb[1]=(BYTE)g;
		rgb[2]=(BYTE)b;
	    }
	    fwrite(v0,sizeof(v0),1,new); fwrite(rgb,sizeof(rgb),1,new);
	    fwrite(v1,sizeof(v1),1,new); fwrite(rgb,sizeof(rgb),1,new);
	    fwrite(v2,sizeof(v2),1,new); fwrite(rgb,sizeof(rgb),1,new);
	}
	if (org!=stdin) fclose(org);

	for (cnt=0; cnt<num; ) {
	    fputc(3,new); fwrite(&cnt,sizeof(cnt),1,new); ++cnt;
			  fwrite(&cnt,sizeof(cnt),1,new); ++cnt;
			  fwrite(&cnt,sizeof(cnt),1,new); ++cnt;
	}
	if (new!=stdout) fclose(new);

	return 0;
}
