
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "stl.h"
#include "cell.h"
#include "sif.h"

#ifndef	BITS
#define BITS	15
#endif

#ifndef	Int
#define Int	int
#endif

static void	Error(msg)
char		*msg;
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

#define IX1(x1)		((x1<=(-0.5))?0:(int)ceil(x1-0.5))
#define IX2(x2,Nx)	((x2>=(double)Nx+0.5)?Nx-1:(int)floor(x2+0.5)-1)

static int	Extent(x0,x1,x2,Nx,ix1,ix2)
double		x0,x1,x2;
int		Nx,*ix1,*ix2;
{
	return (x1<x2)?(*ix1=(x0<x1)?IX1(x0):IX1(x1))<=
		       (*ix2=(x0>x2)?IX2(x0,Nx):IX2(x2,Nx)):
		       (*ix1=(x0<x2)?IX1(x0):IX1(x2))<=
		       (*ix2=(x0>x1)?IX2(x0,Nx):IX2(x1,Nx));
}

typedef	struct Edge {
		Int		x;
		struct Edge	*next;
	} Edge;

#define MF	(double)(1<<BITS)

static void	Print(edge)
Edge		*edge;
{
	while (edge!=NULL) {
	    (void)printf("\t%lf",(double)edge->x/MF); edge=edge->next;
	}
	(void)putchar('\n');
}

static void	Free(edge)
Edge		*edge;
{
	Edge	*next;

	while (edge!=NULL) {
	    next=edge->next; free(edge); edge=next;
	}
}

#define BYTE	unsigned char

typedef struct {
		Edge	*edge[2];
		BYTE	*line;
	} Wall;

#define STR_LEN	1024

int	main(argc,argv)
int	argc;
char	**argv;
{
	double	xO,yO,zO,Sx,Sy,Sz,
		z0,z1,z2,Y0,Y1,y2,C,x0,x1,x2,A,B,z,y1z,y2z,y,c,a,b;
	int	Nx,Ny,Nz,iz,iy,ix,iz1,iz2,iy1,iy2,e,mask,len;
	Wall	**wall,*w;
	BYTE	*line;
	FILE	*file;
	STL	stl;
	float	f[9];
	Edge	**E,*edge,*E1,*E2,*e1,*e2;
	Int	x;
	Cell	**cell;
	char	form[STR_LEN],path[STR_LEN],
		*wmae="wall memory allocation error.",
		*cmae="cell memory allocation error.";

	if (argc<12)
	Error("usage : stl_si STL_files xO yO zO Sx Sy Sz Nx Ny Nz directory");

	xO=atof(argv[argc-10]);
	yO=atof(argv[argc- 9]);
	zO=atof(argv[argc- 8]);

	if ((Sx=atof(argv[argc-7]))<=0.0 ||
	    (Sy=atof(argv[argc-6]))<=0.0 ||
	    (Sz=atof(argv[argc-5]))<=0.0) Error("bad pixel size.");

	if ((Nx=atoi(argv[argc-4]))<=0 ||
	    (Ny=atoi(argv[argc-3]))<=0 ||
	    (Nz=atoi(argv[argc-2]))<=0) Error("bad pixel number.");

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size_t)(size))

	if ((wall=ALLOC(Wall *,Nz))==NULL) Error(wmae);

	for (iz=0; iz<Nz; iz++) {
	    if ((wall[iz]=ALLOC(Wall,Ny))==NULL) Error(wmae);

	    for (w=wall[iz], iy=0; iy<Ny; iy++, w++) {
		w->edge[0]=w->edge[1]=NULL;

		if ((w->line=ALLOC(BYTE,(Nx+7)/8))==NULL) Error(wmae);

		for (line=w->line, ix=(Nx+7)/8; ix>0; ix--, line++) *line=0;
	    }
	}
	while (--argc>10) {
	    if (**(++argv)=='-')
		file=stdin;
	    else
		if ((file=fopen(*argv,"rb"))==NULL) {
		    (void)fputs(*argv,stderr); Error(" : STL file not found.");
		}

	    if (!SetUpSTL(&stl,file)) {
		(void)fputs(*argv,stderr); Error(" : STL header read error.");
	    }
	    while (ScanSTL(&stl,f+0,f+1,f+2,f+3,f+4,f+5,f+6,f+7,f+8)>0) {
		if (!Extent(z0=(f[2]-zO)/Sz,
			    z1=(f[5]-zO)/Sz,
			    z2=(f[8]-zO)/Sz,Nz,&iz1,&iz2) ||
		    !Extent(Y0=(f[1]-yO)/Sy,
			    Y1=(f[4]-yO)/Sy,
			    y2=(f[7]-yO)/Sy,Ny,&iy1,&iy2)) continue;

		C=(Y1-=Y0)*(z2-=z0)-
		  (y2-=Y0)*(z1-=z0); if (C<0.0) e=0; else
				     if (C>0.0) e=1; else continue;

		x0=(f[0]-xO)/Sx; x1=(f[3]-xO)/Sx-x0;
				 x2=(f[6]-xO)/Sx-x0; A=(x1*z2-x2*z1)/C;
						     B=(Y1*x2-y2*x1)/C;
		for (iz=iz1; iz<=iz2; iz++) {
		    z=(double)iz+0.5-z0; y1z=Y1*z; y2z=y2*z;
		    for (w=wall[iz]+(iy=iy1); iy<=iy2; iy++, w++) {
			y=(double)iy+0.5-Y0;
			c=(a=y*z2-y2z)+
			  (b=y1z-y*z1); if ((a<0.0 || b<0.0 || c>C) &&
					    (a>0.0 || b>0.0 || c<C)) continue;

			E=w->edge+e;
			x=(Int)floor((x0+A*y+B*z)*MF+0.5);
			while (*E!=NULL && (*E)->x<x) E=(&((*E)->next));

			if (*E==NULL || (*E)->x>x) {
			    if ((edge=ALLOC(Edge,1))==NULL)
				Error("edge memory allocation error.");

			    edge->x=x; edge->next=(*E); *E=edge;
			}
		    }
		}
	    }
	    if (file!=stdin) (void)fclose(file);

	    for (iz=0; iz<Nz; iz++)
		for (w=wall[iz], iy=0; iy<Ny; iy++, w++) {
		    E1=w->edge[0];
		    E2=w->edge[1]; if (E1==NULL && E2==NULL) continue;

		    for (e1=E1,
			 e2=E2; e1!=NULL &&
				e2!=NULL; e1=e1->next,
					  e2=e2->next) if (e1->x>e2->x) break;

		    if (e1!=NULL || e2!=NULL) {
			(void)printf("%d\t%d\n",iz,iy); Print(E1);
							Print(E2);
		    }
		    else
			for (e1=E1,
			     e2=E2; e1!=NULL; e1=e1->next,
					      e2=e2->next) {
			    if ((x1=(double)e1->x/MF)>(double)Nx-0.5) break;

			    if ((x2=(double)e2->x/MF)>=0.5) {
				ix=IX1(x1); line=w->line+ix/8; mask=1<<ix%8;
				for (ix=IX2(x2,Nx)-ix; ix>=0; ix--) {
				    *line|=(BYTE)mask;
				    if ((mask<<=1)==256) {
					++line; mask=1;
				    }
				}
			    }
			}

		    Free(E1);
		    Free(E2); w->edge[0]=w->edge[1]=NULL;
		}
	}
	if ((cell=ALLOC(Cell *,Ny))==NULL) Error(cmae);

	for (iy=0; iy<Ny; iy++)
	    if ((cell[iy]=ALLOC(Cell,Nx))==NULL) Error(cmae);

	len=1; for (iz=10; iz<Nz; iz*=10) ++len;
	(void)sprintf(form,"%s/%%0%dd.tif",argv[argc],len);

	for (iz=0; iz<Nz; iz++) {
	    for (w=wall[iz], iy=0; iy<Ny; iy++, w++) {
		line=w->line; mask=1;
		for (ix=0; ix<Nx; ix++) {
		    cell[iy][ix]=(*line&mask)?1:0;
		    if ((mask<<=1)==256) {
			++line; mask=1;
		    }
		}
	    }
	    (void)sprintf(path,form,iz);
	    StoreImageFile(path,Nx,Ny,1,cell,NULL);
	}
	return 0;
}
