
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define STR_LEN	256

static char	line[STR_LEN];

static void	Scan(file)
FILE		*file;
{
	char	str[STR_LEN];

	for (;;) {
	    if (fgets(line,STR_LEN,file)==NULL)
		Error("unexpected end of file.");

	    if (line[strlen(line)-1]!='\n') Error("missing end of line.");

	    if (sscanf(line,"%s",str)==1 && strcmp(str,"comment")) return;
	}
}

static int	CheckElement(file,element)
FILE		*file;
char		*element;
{
	char	str[3][STR_LEN];
	int	number;

	Scan(file);
	if (sscanf(line,"%s %s %s",str[0],str[1],str[2])!=3 ||
	    strcmp(str[0],"element") ||
	    strcmp(str[1],element) ||
	    sscanf(str[2],"%d",&number)!=1 ||
	    number<=0) {
	    fprintf(stderr,"bad element specification for %s\n",element);
	    exit(1);
	}
	return number;
}

static void	CheckProperty(file,type,name,element)
FILE		*file;
char		*type,*name,*element;
{
	char	str[3][STR_LEN];

	Scan(file);
	if (sscanf(line,"%s %s %s",str[0],str[1],str[2])!=3 ||
	    strcmp(str[0],"property") ||
	    strcmp(str[1],type) ||
	    strcmp(str[2],name)) {
	    fprintf(stderr,"bad property specification for %s\n",element);
	    exit(1);
	}
}

static size_t	freadReverse(ptr,size,items,file)
unsigned char	*ptr;
size_t		size,items;
FILE		*file;
{
	size_t		item,size1=size-1,read=fread(ptr,size,items,file);
	unsigned char	*head,*tail,uc;

	for (item=0; item<read; item++) {
	    for (tail=(head=ptr)+size1; head<tail; head++, tail--) {
		uc=(*head); *head=(*tail); *tail=uc;
	    }
	    ptr+=size;
	}
	return read;
}

static size_t	fwriteReverse(ptr,size,items,file)
unsigned char	*ptr;
size_t		size,items;
FILE		*file;
{
	size_t		written,size1=size-1;
	unsigned char	*uc;

	for (written=0; written<items; written++) {
	    for (uc=ptr+size1; uc>=ptr; uc--)
		if (fputc(*uc,file)!=*uc) return written;

	    ptr+=size;
	}
	return written;
}

#define MA(noe,ptr)	malloc((noe)*sizeof(*(ptr)))

int	main(argc,argv)
int	argc;
char	**argv;
{
	FILE	*zcp,*stl;
	char	str[5][STR_LEN];
	int	vertices,faces,face,vertex,idx,R,G,B;
	float	*XYZ,xyz[12],*f,*F;
	double	x10,x20,y10,y20,z10,z20,nx,ny,nz,n;
	short	flag;

	size_t	(*fRead)(),(*fWrite)();
{
	short	s=0x1234;
	char	*c=(char *)&s;

	if (*c==(char)(s&0xff)) {
	    fRead=fread; fWrite=fwrite;
	}
	else {
	    fRead=freadReverse; fWrite=fwriteReverse;
	}
}
	if (argc!=3) Error("usage : zcp_stl ZCP STL");

	if (argv[1][0]=='-')
	    zcp=stdin;
	else
	    if ((zcp=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (fgetc(zcp)!='p' ||
	    fgetc(zcp)!='l' ||
	    fgetc(zcp)!='y') Error("bad magic number.");

	Scan(zcp);
	if (sscanf(line,"%s %s %s",str[0],str[1],str[2])!=3 ||
	    strcmp(str[0],"format") ||
	    strcmp(str[1],"binary_little_endian") ||
	    strcmp(str[2],"1.0"))
	    Error("bad format specification.");

	vertices=CheckElement(zcp,"vertex");

	CheckProperty(zcp,"float","x","vertex");
	CheckProperty(zcp,"float","y","vertex");
	CheckProperty(zcp,"float","z","vertex");

	faces=CheckElement(zcp,"face");

	CheckProperty(zcp,"uchar","red","face");
	CheckProperty(zcp,"uchar","green","face");
	CheckProperty(zcp,"uchar","blue","face");

	Scan(zcp);
	if (sscanf(line,"%s %s %s %s %s",
			str[0],str[1],str[2],str[3],str[4])!=5 ||
	    strcmp(str[0],"property") ||
	    strcmp(str[1],"list") ||
	    strcmp(str[2],"uchar") ||
	    strcmp(str[3],"int") ||
	    strcmp(str[4],"vertex_indices"))
	    Error("bad vertex index specification for face.");

	Scan(zcp);
	if (sscanf(line,"%s",str[0])!=1 ||
	    strcmp(str[0],"end_header")) Error("missing end of header.");

	if ((XYZ=(float *)MA(vertices*3,XYZ))==NULL)
	    Error("vertex memory allocation error.");

	if ((*fRead)(XYZ,sizeof(*XYZ),vertices*3,zcp)!=vertices*3)
	    Error("vertex data read error.");

	if (argv[2][0]=='-')
	    stl=stdout;
	else
	    if ((stl=fopen(argv[2],"wb"))==NULL) Error("file not stored.");

	for (idx=0; idx<80; idx++) fputc(0,stl);

	(*fWrite)(&faces,sizeof(faces),1,stl);

	for (face=0; face<faces; face++) {
	    if ((R=fgetc(zcp))==EOF ||
		(G=fgetc(zcp))==EOF ||
		(B=fgetc(zcp))==EOF) Error("color data read error.");

	    if (fgetc(zcp)!=3) Error("bad number of vertices per face.");

	    for (f=xyz+3, idx=0; idx<3; idx++, f+=3) {
		if ((*fRead)(&vertex,sizeof(vertex),1,zcp)!=1)
		    Error("face data read error.\n");

		if (vertex<0 || vertex>=vertices) Error("bad vertex index.");

		F=XYZ+vertex*3; f[0]=F[0]; f[1]=F[1]; f[2]=F[2];
	    }
	    x10=xyz[6]-xyz[3]; x20=xyz[ 9]-xyz[3];
	    y10=xyz[7]-xyz[4]; y20=xyz[10]-xyz[4];
	    z10=xyz[8]-xyz[5]; z20=xyz[11]-xyz[5];
	    nx=y10*z20-z10*y20;
	    ny=z10*x20-x10*z20;
	    nz=x10*y20-y10*x20; n=sqrt(nx*nx+ny*ny+nz*nz);
	    xyz[0]=nx/n;
	    xyz[1]=ny/n;
	    xyz[2]=nz/n; (*fWrite)(xyz,sizeof(*xyz),12,stl);

	    flag=(unsigned short)(0x8000|((R>>3)<<10)|((G>>3)<<5)|(B>>3));
	    (*fWrite)(&flag,sizeof(flag),1,stl);
	}
	if (zcp!=stdin ) fclose(zcp);
	if (stl!=stdout) fclose(stl);

	return 0;
}
